--[[
Name: RatingBuster_AlwaysBuffed.lua
Description: Enables RatingBuster to calculate selected buff effects even if you don't really have them
Revision: $Revision: 11 $
Author: Whitetooth
Email: hotdogee [at] gmail [dot] com
LastUpdate: $Date: 2010-10-19 17:02:55 +0000 (Tue, 19 Oct 2010) $
]]
-- version check, only works with RatingBuster r279 and LibStatLogic-1.2
if tonumber((gsub(RatingBuster.version, "%d+%.%d+%.%d+ %(r(%d+)%)", "%1"))) < 279 then
  error("RatingBuster_AlwaysBuffed requires RatingBuster r279 or higher")
  return
end
local StatLogic, sl_ver = LibStub("LibStatLogic-1.2")

-- Locale --
local L
L = LibStub("AceLocale-3.0"):NewLocale("RatingBuster_AlwaysBuffed", "enUS", true)
L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"] = true
L["$class Self Buffs"] = true -- $class will be replaced with localized player class
L["Raid Buffs"] = true
L = LibStub("AceLocale-3.0"):NewLocale("RatingBuster_AlwaysBuffed", "zhTW")
if L then
L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"] = "指定常駐buff，就算身上沒有buff，RatingBuster也會當成有來計算"
L["$class Self Buffs"] = "$class個人Buff"
L["Raid Buffs"] = "團隊Buff"
end
L = LibStub("AceLocale-3.0"):NewLocale("RatingBuster_AlwaysBuffed", "zhCN")
if L then
L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"] = "指定常驻buff，就算身上没有buff，RatingBuster也会当成有来计算"
L["$class Self Buffs"] = "$class个人Buff"
L["Raid Buffs"] = "团队Buff"
end
L = LibStub("AceLocale-3.0"):NewLocale("RatingBuster_AlwaysBuffed", "koKR")
if L then
-- L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"] = ""
-- L["$class Self Buffs"] = ""
-- L["Raid Buffs"] = ""
end
L = LibStub("AceLocale-3.0"):NewLocale("RatingBuster_AlwaysBuffed", "deDE")
if L then
L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"] = "Erlaubt RatingBuster gewählte Buffs zu berechnen, auch wenn du diese nicht wirklich hast"
L["$class Self Buffs"] = "Eigenbuffs"
L["Raid Buffs"] = "Raidbuffs"
end
L = LibStub("AceLocale-3.0"):NewLocale("RatingBuster_AlwaysBuffed", "frFR")
if L then
-- L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"] = ""
-- L["$class Self Buffs"] = ""
-- L["Raid Buffs"] = ""
end
L = LibStub("AceLocale-3.0"):NewLocale("RatingBuster_AlwaysBuffed", "ruRU")
if L then
-- L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"] = ""
-- L["$class Self Buffs"] = ""
-- L["Raid Buffs"] = ""
end
L = LibStub("AceLocale-3.0"):NewLocale("RatingBuster_AlwaysBuffed", "esES")
if L then
-- L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"] = ""
-- L["$class Self Buffs"] = ""
-- L["Raid Buffs"] = ""
end

-- Libraries --
L = LibStub("AceLocale-3.0"):GetLocale("RatingBuster_AlwaysBuffed")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")
local AceDB = LibStub("AceDB-3.0")
local RBL = LibStub("AceLocale-3.0"):GetLocale("RatingBuster")


local rba = RatingBuster:NewModule("RatingBuster_AlwaysBuffed", "AceConsole-3.0", "AceEvent-3.0")

local _, playerClass = UnitClass("player")
local _, playerRace = UnitRace("player")
local wowBuildNo = tonumber((select(2, GetBuildInfo())))
local toc = tonumber((select(4, GetBuildInfo())))
local GetTalentInfo = GetTalentInfo
local GetSpellInfo = GetSpellInfo
local loadstring = loadstring
local pairs = pairs
local ipairs = ipairs

-- statmod tables
local statMods = {
	["ADD_AP_MOD_INT"] = true,
	["ADD_AP_MOD_STA"] = true,
	["ADD_AP_MOD_ARMOR"] = true,
	["ADD_AP_MOD_SPELL_DMG"] = true,
	["ADD_PARRY_RATING_MOD_STR"] = true,
	["ADD_SPELL_CRIT_RATING_MOD_SPI"] = true,
	["ADD_RANGED_AP_MOD_INT"] = true,
	["ADD_ARMOR_MOD_INT"] = true,
	["ADD_SPELL_DMG_MOD_AP"] = true,
	["ADD_SPELL_DMG_MOD_STR"] = true,
	["ADD_SPELL_DMG_MOD_STA"] = true,
	["ADD_SPELL_DMG_MOD_INT"] = true,
	["ADD_SPELL_DMG_MOD_INT"] = true,
	["ADD_SPELL_DMG_MOD_SPI"] = true,
	["ADD_SPELL_DMG_MOD_PET_STA"] = true,
	["ADD_SPELL_DMG_MOD_PET_INT"] = true,
	["ADD_SPELL_DMG_MOD_MANA"] = true,
	["ADD_HEAL_MOD_AP"] = true,
	["ADD_HEAL_MOD_STR"] = true,
	["ADD_HEAL_MOD_AGI"] = true,
	["ADD_HEAL_MOD_STA"] = true,
	["ADD_HEAL_MOD_INT"] = true,
	["ADD_HEAL_MOD_SPI"] = true,
	["ADD_HEAL_MOD_MANA"] = true,
	["ADD_MANA_REG_MOD_INT"] = true,
	["ADD_MANA_REG_MOD_MANA"] = true,
	["ADD_MANA_REG_MOD_NORMAL_MANA_REG"] = true,
	["ADD_PET_STA_MOD_STA"] = true,
	["ADD_PET_INT_MOD_INT"] = true,
	["MOD_ARMOR"] = true,
	["MOD_HEALTH"] = true,
	["MOD_MANA"] = true,
	["MOD_STR"] = true,
	["MOD_AGI"] = true,
	["MOD_STA"] = true,
	["MOD_INT"] = true,
	["MOD_SPI"] = true,
	["MOD_BLOCK_VALUE"] = true,
	["MOD_AP"] = true,
	["MOD_RANGED_AP"] = true,
	["MOD_SPELL_DMG"] = true,
	["MOD_HEAL"] = true,
}

-- Generate AuraInfo --
AuraInfo = setmetatable({}, { __index = function(self, n)
  self[n] = {rank = 1, stack = 1}
  return self[n]
end })
local function GenerateAuraInfo() -- run OnInitialize to make sure the newest version of LibStatLogic is loaded
  for class, tables in pairs(StatLogic.StatModTable) do
    for modName, mods in pairs(tables) do
      -- filter out uninterested stat mods
      if statMods[modName] then
        for key, mod in pairs(mods) do
          if mod.old and wowBuildNo >= mod.old then
          elseif mod.new and wowBuildNo < mod.new then
          elseif mod.buff then -- if we got a buff
            if GetSpellInfo(mod.buff) then
              if not mod.tab then -- not a talent, so the rank is the buff rank
                AuraInfo[(GetSpellInfo(mod.buff))].rank = #(mod.rank)
              end
              AuraInfo[(GetSpellInfo(mod.buff))].stack = mod.buffStack or 1
            end
          end
        end
      end
    end
  end
end

-- AceDB setup --
local settings -- = self.db.profile
local defaults = {
  profile = {
    ['*'] = false
  }
}

-- Build options table --
local options = {
  type = "group",
  name = "RatingBuster AlwaysBuffed",
  get = function(info)
    return settings[info[#info]]
  end,
  set = function(info, v)
    settings[info[#info]] = v
    RatingBuster:ClearCache()
  end,
  args = {
    main = {
      type = "group",
      name = "main",
      args = {
        description = {
					type = "description",
          name = L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"],
          order = 1,
        },
        description2 = {
					type = "description",
          name = " ",
          order = 2,
        },
        class = {
					type = "group",
					dialogInline = true,
          name = gsub(L["$class Self Buffs"], "$class", (UnitClass("player"))),
          order = 5,
					args = {
          },
        },
        general = {
					type = "group",
					dialogInline = true,
          name = L["Raid Buffs"],
          order = 6,
					args = {
          },
        },
      },
    },
  }
}

-- /dump select(11, UnitAura("player", 1))
local function SetupOptions()
  -- tooltip to scan buff desc
	local tip = CreateFrame("GameTooltip", "RatingBuster_AlwaysBuffedTooltip", nil, "GameTooltipTemplate")
	tip:SetOwner(UIParent, "ANCHOR_NONE")
	for i = 1, 5 do
		tip[i] = _G["RatingBuster_AlwaysBuffedTooltipTextLeft"..i]
		if not tip[i] then
			tip[i] = tip:CreateFontString()
			tip:AddFontStrings(tip[i], tip:CreateFontString())
			_G["RatingBuster_AlwaysBuffedTooltipTextLeft"..i] = tip[i]
		end
	end
  -- class buffs
  local r, g, b, name, _, icon
  for modName, mods in pairs(StatLogic.StatModTable[playerClass]) do
    if statMods[modName] then -- filter out uninterested stat mods
      for key, mod in pairs(mods) do
        if mod.old and wowBuildNo >= mod.old then
        elseif mod.new and wowBuildNo < mod.new then
        elseif mod.buff then -- if we got a buff
          name, _, icon = GetSpellInfo(mod.buff)
          -- if not name or not icon then
            -- print(name, icon, mod.buff)
          -- end
          if name and icon then
            options.args.main.args.class.args[name] = {
              ["type"] = 'toggle',
              ["name"] = "|T"..icon..":25:25:-2:0|t"..name,
            }
            tip:ClearLines()
            tip:SetHyperlink("spell:"..mod.buff)
            for i = 2, 5 do
              r, g, b = tip[i]:GetTextColor()
              if b < 0.2 then
                options.args.main.args.class.args[name].desc = tip[i]:GetText()
                break
              end
            end
          end
        end
      end
    end
  end
  -- general buffs
  for modName, mods in pairs(StatLogic.StatModTable["ALL"]) do
    if statMods[modName] then -- filter out uninterested stat mods
      for key, mod in pairs(mods) do
        if mod.old and wowBuildNo >= mod.old then
        elseif mod.new and wowBuildNo < mod.new then
        elseif mod.buff then -- if we got a buff
          local name, _, icon = GetSpellInfo(mod.buff)
          if name and icon then
            options.args.main.args.general.args[name] = {
              ["type"] = 'toggle',
              ["name"] = "|T"..icon..":25:25:-2:0|t"..name,
            }
            tip:ClearLines()
            tip:SetHyperlink("spell:"..mod.buff)
            for i = 2, 5 do
              r, g, b = tip[i]:GetTextColor()
              if b == 0 then
                options.args.main.args.general.args[name].desc = tip[i]:GetText()
                break
              end
            end
          end
        end
      end
    end
  end
	-- Inject profile options
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(rba.db)
	
	-- Add dual-spec support
	local LibDualSpec = LibStub("LibDualSpec-1.0", true)
	if LibDualSpec then
		LibDualSpec:EnhanceDatabase(rba.db, "RatingBuster_AlwaysBuffed")
		LibDualSpec:EnhanceOptions(options.args.profile, rba.db)
	end
	
	-- Register options table
	AceConfig:RegisterOptionsTable("RatingBuster_AlwaysBuffed", options)
	
	-- The ordering here matters, it determines the order in the Blizzard Interface Options
	AceConfigDialog:AddToBlizOptions("RatingBuster_AlwaysBuffed", "RatingBuster AlwaysBuffed", nil, "main")
	AceConfigDialog:AddToBlizOptions("RatingBuster_AlwaysBuffed", RBL["Profiles"], "RatingBuster AlwaysBuffed", "profile")
end


-- Initializations --
function rba:OnProfileChanged(event, database, newProfileKey)
	-- this is called every time our profile changes (after the change)
	settings = database.profile
	--clearCache()
end

-- OnInitialize(name) called at ADDON_LOADED
function rba:OnInitialize()
	-- Create DB
	self.db = AceDB:New("RatingBuster_AlwaysBuffedDB", defaults)
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	settings = self.db.profile

	GenerateAuraInfo()
  
  -- Only generate the options when InterfaceOptionsFrame OnShow
  local f = CreateFrame('Frame', nil, InterfaceOptionsFrame)
  f:SetScript('OnShow', function(self)
    SetupOptions()
    self:SetScript('OnShow', nil)
  end)
end

-- Modified GetStatMod --
local function RBA_PlayerHasAura(aura)
  if settings[aura] == true then
    return true
  else
    return StatLogic.PlayerHasAura(aura)
  end
end

local function RBA_GetPlayerAuraRankStack(aura)
  if settings[aura] == true then
    return AuraInfo[aura].rank or 1, AuraInfo[aura].stack or 1
  else
    return StatLogic.GetPlayerAuraRankStack(aura)
  end
end

local buffGroup = {}
local function GetStatMod(self, stat, school, talentGroup)
	local statModInfo = StatLogic.StatModInfo[stat]
  if not statModInfo then error(stat) end
	local mod = statModInfo.initialValue
	-- if school is required for this statMod but not given
	if statModInfo.school and not school then return mod end
  wipe(buffGroup)
	-- Class specific mods
	if type(StatLogic.StatModTable[playerClass][stat]) == "table" then
		for _, case in ipairs(StatLogic.StatModTable[playerClass][stat]) do
			local ok = true
			if school and not case[school] then ok = nil end
			if ok and case.newtoc and toc < case.newtoc then ok = nil end
			if ok and case.oldtoc and toc >= case.oldtoc then ok = nil end
			if ok and case.new and wowBuildNo < case.new then ok = nil end
			if ok and case.old and wowBuildNo >= case.old then ok = nil end
			if ok and case.condition and not loadstring("return "..case.condition)() then ok = nil end
			if ok and case.buffName and not RBA_PlayerHasAura(case.buffName) then ok = nil end
			if ok and case.buff2Name and not RBA_PlayerHasAura(case.buff2Name) then ok = nil end
			if ok and case.stance and case.stance ~= StatLogic.GetStanceIcon() then ok = nil end
			if ok and case.glyph and not StatLogic.PlayerHasGlyph(case.glyph, talentGroup) then ok = nil end
			if ok and case.enchant and not StatLogic.SlotHasEnchant(case.enchant, case.slot) then ok = nil end
			if ok and case.itemset and ((not StatLogic:PlayerHasItemSet(case.itemset[1])) or StatLogic:PlayerHasItemSet(case.itemset[1]) < case.itemset[2]) then ok = nil end
			if ok then
				local r, _
				local s = 1
				-- if talent field
				if case.tab and case.num then
					_, _, _, _, r = GetTalentInfo(case.tab, case.num, nil, nil, talentGroup)
					if case.buffName and case.buffStack then
						_, s = RBA_GetPlayerAuraRankStack(case.buffName) -- Gets buff stack count, but use talent as rank
					end
				-- no talent but buff is given
				elseif case.buffName then
					r, s = RBA_GetPlayerAuraRankStack(case.buffName)
					if not case.buffStack then
						s = 1
					end
				-- no talent but all other given conditions are statisfied
				else--if case.condition or case.stance then
					r = 1
				end
				if r and r ~= 0 and case.rank[r] then
					if statModInfo.initialValue == 0 then
            if not case.group then
              mod = mod + case.rank[r] * s
            elseif not buffGroup[case.group] then -- this mod is part of a group, but not seen before
              mod = mod + case.rank[r] * s
              buffGroup[case.group] = case.rank[r] * s
            elseif (case.rank[r] * s) > buffGroup[case.group] then -- seen before and this one is better, do upgrade
              mod = mod + case.rank[r] * s - buffGroup[case.group]
              buffGroup[case.group] = case.rank[r] * s
            else -- seen before but not better, do nothing
            end
					else
            if not case.group then
              mod = mod * (case.rank[r] * s + 1)
            elseif not buffGroup[case.group] then -- this mod is part of a group, but not seen before
              mod = mod * (case.rank[r] * s + 1)
              buffGroup[case.group] = (case.rank[r] * s + 1)
            elseif (case.rank[r] * s + 1) > buffGroup[case.group] then -- seen before and this one is better, do upgrade
              mod = mod * (case.rank[r] * s + 1) / buffGroup[case.group]
              buffGroup[case.group] = (case.rank[r] * s + 1)
            else -- seen before but not better, do nothing
            end
					end
				end
			end
		end
	end
	-- Non class specific mods
	if type(StatLogic.StatModTable["ALL"][stat]) == "table" then
		for _, case in ipairs(StatLogic.StatModTable["ALL"][stat]) do
			local ok = true
			if school and not case[school] then ok = nil end
			if ok and case.newtoc and toc < case.newtoc then ok = nil end
			if ok and case.oldtoc and toc >= case.oldtoc then ok = nil end
			if ok and case.new and wowBuildNo < case.new then ok = nil end
			if ok and case.old and wowBuildNo >= case.old then ok = nil end
			if ok and case.condition and not loadstring("return "..case.condition)() then ok = nil end
			if ok and case.buffName and not RBA_PlayerHasAura(case.buffName) then ok = nil end
			if ok and case.stance and case.stance ~= StatLogic.GetStanceIcon() then ok = nil end
			if ok and case.race and case.race ~= playerRace then ok = nil end
			if ok and case.meta and not StatLogic:IsMetaGemActive(case.meta) then ok = nil end
			if ok then
				local r, _
				local s = 1
				-- if talent field
				if case.tab and case.num then
					_, _, _, _, r = GetTalentInfo(case.tab, case.num, nil, nil, talentGroup)
					if case.buffName and case.buffStack then
						_, s = RBA_GetPlayerAuraRankStack(case.buffName) -- Gets buff rank and stack count
					end
				-- no talent but buff is given
				elseif case.buffName then
					r, s = RBA_GetPlayerAuraRankStack(case.buffName)
					if not case.buffStack then
						s = 1
					end
				-- no talent but all other given conditions are statisfied
				else--if case.condition or case.stance then
					r = 1
				end
				if r and r ~= 0 and case.rank[r] then
					if statModInfo.initialValue == 0 then
            if not case.group then
              --print(1, stat, case.group, case.buffName)
              mod = mod + case.rank[r] * s
            elseif not buffGroup[case.group] then -- this mod is part of a group, but not seen before
              --print(2, stat, case.group, case.buffName)
              mod = mod + case.rank[r] * s
              buffGroup[case.group] = case.rank[r] * s
            elseif (case.rank[r] * s) > buffGroup[case.group] then -- seen before and this one is better, do upgrade
              --print(3, stat, case.group, case.buffName)
              mod = mod + case.rank[r] * s - buffGroup[case.group]
              buffGroup[case.group] = case.rank[r] * s
            else -- seen before but not better, do nothing
            end
					else
            if not case.group then
              --print(1, stat, case.group, case.buffName)
              mod = mod * (case.rank[r] * s + 1)
            elseif not buffGroup[case.group] then -- this mod is part of a group, but not seen before
              --print(2, stat, case.group, case.buffName)
              mod = mod * (case.rank[r] * s + 1)
              buffGroup[case.group] = (case.rank[r] * s + 1)
            elseif (case.rank[r] * s + 1) > buffGroup[case.group] then -- seen before and this one is better, do upgrade
              --print(3, stat, case.group, case.buffName)
              mod = mod * (case.rank[r] * s + 1) / buffGroup[case.group]
              buffGroup[case.group] = (case.rank[r] * s + 1)
            else -- seen before but not better, do nothing
            end
					end
				end
			end
		end
	end
	return mod + statModInfo.finalAdjust
end
  
-- Replace RatingBuster's GetStatMod
RatingBuster.GetStatMod = GetStatMod